@@ -75,17 +75,18 @@ inline RuleType& operator++(RuleType& x) {
 
 //! Supported edge types.
 /*!
- * Currently, we distinguish three types of edges:
+ * Currently, we distinguish four types of edges:
  *  - a NORMAL_EDGE stipulates an implication between body and head,
  *    i.e. tableau-rules FTA and BFA for atoms.
  *  - a CHOICE_EDGE only stipulates a support.
  *  - a GAMMA_EDGE is like a NORMAL_EDGE that is only considered during
  *    nogood creation but ignored in the dependency graph.
+ *  - a GAMMA_CHOICE_EDGE is like a GAMMA_EDGE but only stipulates a support.
  * The head of a rule is either an atom or a disjunction.
  */
 struct PrgEdge {
 	//! Type of edge.
-	enum EdgeType{ NORMAL_EDGE = 0, GAMMA_EDGE = 1, CHOICE_EDGE = 2};
+	enum EdgeType{ NORMAL_EDGE = 0, GAMMA_EDGE = 1, CHOICE_EDGE = 2, GAMMA_CHOICE_EDGE = 3};
 	//! Type of adjacent node.
 	enum NodeType{ BODY_NODE   = 0, ATOM_NODE  = 1, DISJ_NODE   = 2};
 
