@@ -22,6 +22,7 @@
 #include <clasp/cb_enumerator.h>
 #include <clasp/dependency_graph.h>
 #include <clasp/minimize_constraint.h>
+#include <clasp/unfounded_check.h>
 #include <clasp/util/timer.h>
 #include <clasp/util/atomic.h>
 #include <cstdio>
@@ -36,54 +37,39 @@ namespace Clasp {
 // ClaspConfig
 /////////////////////////////////////////////////////////////////////////////////////////
 ClaspConfig::ClaspConfig() : tester_(0) { }
-ClaspConfig::~ClaspConfig() {
-	setSolvers(1);
-	delete tester_;
-}
+ClaspConfig::~ClaspConfig() { delete tester_; }
 void ClaspConfig::reset() {
-	BasicSatConfig::reset();
-	solve     = SolveOptions();
-	enumerate = EnumOptions();
-	asp       = AspOptions();
 	if (tester_) { tester_->reset(); }
-	setSolvers(1);
+	BasicSatConfig::reset();
+	solve = SolveOptions();
+	asp   = AspOptions();
 }
 BasicSatConfig* ClaspConfig::addTesterConfig() {
 	if (!tester_) { tester_ = new BasicSatConfig(); }
 	return tester_;
 }
-void ClaspConfig::setSolvers(uint32 num) {
-	if (!num) { num = 1; }
-#if WITH_THREADS
-	solve.algorithm.threads = num;
-#endif
-	if (num < numSolver()) { BasicSatConfig::resize(num, std::min(num, numSearch())); }
-	if (testerConfig() && num < testerConfig()->numSolver()) { testerConfig()->resize(num, std::min(num, testerConfig()->numSearch())); }
-}
 void ClaspConfig::prepare(SharedContext& ctx) {
+	BasicSatConfig::prepare(ctx);
 	uint32 numS = solve.numSolver();
-	if (numS > solve.recommendedSolvers()) {
-		ctx.report(warning(Event::subsystem_facade, clasp_format_error("Oversubscription: #Threads=%u exceeds logical CPUs=%u.", numS, solve.recommendedSolvers())));
-	}
 	if (numS > solve.supportedSolvers()) {
 		ctx.report(warning(Event::subsystem_facade, "Too many solvers."));
 		numS = solve.supportedSolvers();
 	}
-	if (numS > 1 && !solve.defaultPortfolio()) {
-		for (uint32 i = 0, sWarn = 0; i != numS; ++i) {
-			if (solver(i).optStrat >= SolverStrategies::opt_unsat) {
-				if (++sWarn == 1) { ctx.report(warning(Event::subsystem_facade, "Splitting: Disabling unsat-core based optimization!")); }
-				addSolver(i).optStrat = SolverStrategies::opt_dec;
-			}
-		}
+	if (numS > solve.recommendedSolvers()) {
+		ctx.report(warning(Event::subsystem_facade, clasp_format_error("Oversubscription: #Threads=%u exceeds logical CPUs=%u.", numS, solve.recommendedSolvers())));
 	}
-	if (std::abs(enumerate.numModels) != 1) { satPre.mode = SatPreParams::prepro_preserve_models; }
-	setSolvers(numS);
-	BasicSatConfig::prepare(ctx);
-	ctx.setConcurrency(solve.numSolver());
-	for (uint32 i = 1; i != ctx.concurrency(); ++i) {
-		if (!ctx.hasSolver(i)) { ctx.addSolver(); }
+	if (std::abs(solve.numModels) != 1) { satPre.mode = SatPreParams::prepro_preserve_models; }
+	solve.setSolvers(numS);
+	ctx.setConcurrency(solve.numSolver(), SharedContext::mode_resize);
+}
+bool ClaspConfig::addPost(Solver& s) const {
+	bool ok = true;
+	if (s.sharedContext() && s.sharedContext()->sccGraph.get()) {
+		DefaultUnfoundedCheck* pp = static_cast<DefaultUnfoundedCheck*>(s.getPost(PostPropagator::priority_reserved_ufs));
+		if (pp) { pp->setReasonStrategy(static_cast<DefaultUnfoundedCheck::ReasonStrategy>(solver(s.id()).loopRep)); }
+		else    { ok = s.addPost(new DefaultUnfoundedCheck(static_cast<DefaultUnfoundedCheck::ReasonStrategy>(solver(s.id()).loopRep))); }
 	}
+	return ok && BasicSatConfig::addPost(s);
 }
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspFacade::SolveImpl/SolveStrategy/AsyncResult
@@ -141,7 +127,7 @@ struct ClaspFacade::SolveImpl {
 	bool                      solving()   const  { return active && active->running();   }
 	const Model*              lastModel() const  { return en.get() ? &en->lastModel() : 0; }
 	const SharedMinimizeData* minimizer() const  { return en.get() ? en->minimizer() : 0; } 
-	bool                      optimize()  const  { return en.get() && en->optimize(); }
+	Enumerator*               enumerator()const  { return en.get(); }
 	int                       modelType() const  { return en.get() ? en->modelType() : 0; }
 	EnumPtr        en;
 	AlgoPtr        algo;
@@ -286,11 +272,15 @@ void ClaspFacade::init(ClaspConfig& config, bool discard) {
 	if (discard) { discardProblem(); }
 	ctx.setConfiguration(0, false); // force reload of configuration once done
 	config_ = &config;
-	SolveImpl::EnumPtr e(config.enumerate.createEnumerator());
+	if (config_->solve.enumMode == EnumOptions::enum_dom_record && config_->solver(0).heuId != Heuristic_t::heu_domain) {
+		ctx.report(warning(Event::subsystem_facade, "Reasoning mode requires domain heuristic and is ignored!"));
+		config_->solve.enumMode = EnumOptions::enum_auto;
+	}
+	SolveImpl::EnumPtr e(config.solve.createEnumerator(config.solve));
 	if (e.get() == 0) { e = EnumOptions::nullEnumerator(); }
 	if (config.solve.numSolver() > 1 && !e->supportsParallel()) {
 		ctx.report(warning(Event::subsystem_facade, "Selected reasoning mode implies #Threads=1."));
-		config.setSolvers(1);
+		config.solve.setSolvers(1);
 	}
 	ctx.setConfiguration(&config, false); // prepare and apply config
 	if (program() && lpStats_.get()) {
@@ -304,49 +294,57 @@ void ClaspFacade::init(ClaspConfig& config, bool discard) {
 	if (discard) { startStep(0); }
 }
 
-void ClaspFacade::initBuilder(ProgramBuilder* in, bool incremental) {
+void ClaspFacade::initBuilder(ProgramBuilder* in) {
 	builder_ = in;
 	assume_.clear();
 	builder_->startProgram(ctx);
-	if (incremental) { 
-		ctx.requestStepVar();
+}
+
+ProgramBuilder& ClaspFacade::start(ClaspConfig& config, ProblemType t) {
+	if      (t == Problem_t::SAT) { return startSat(config); }
+	else if (t == Problem_t::PB)  { return startPB(config);  }
+	else if (t == Problem_t::ASP) { return startAsp(config); }
+	else                          { throw std::domain_error("Unknown problem type!"); }
+}
+
+SatBuilder& ClaspFacade::startSat(ClaspConfig& config) {
+	init(config, true);
+	initBuilder(new SatBuilder(config.solve.maxSat));
+	return static_cast<SatBuilder&>(*builder_.get());
+}
+
+PBBuilder& ClaspFacade::startPB(ClaspConfig& config) {
+	init(config, true);
+	initBuilder(new PBBuilder());
+	return static_cast<PBBuilder&>(*builder_.get());
+}
+
+Asp::LogicProgram& ClaspFacade::startAsp(ClaspConfig& config, bool enableUpdates) {
+	init(config, true);
+	Asp::LogicProgram* p = new Asp::LogicProgram();
+	lpStats_             = new Asp::LpStats;
+	p->accu              = lpStats_.get();
+	initBuilder(p);
+	p->setOptions(config.asp);
+	p->setNonHcfConfiguration(config.testerConfig());
+	if (enableUpdates) { enableProgramUpdates(); }
+	return *p;
+}
+
+bool ClaspFacade::enableProgramUpdates() {
+	CLASP_ASSERT_CONTRACT_MSG(program(), "Program was already released!");
+	CLASP_ASSERT_CONTRACT(!solving() && !program()->frozen());
+	if (!accu_.get()) {
 		builder_->updateProgram();
+		ctx.requestStepVar();
 		solve_->algo->enableInterrupts();
 		accu_ = new Summary();
 		accu_->init(*this);
 		accu_->step = UINT32_MAX;
 	}
+	return lpStats_.get() != 0; // currently only ASP supports program updates
 }
 
-ProgramBuilder& ClaspFacade::start(ClaspConfig& config, ProblemType t, bool allowUpdate) {
-	if      (t == Problem_t::SAT) { return startSat(config, allowUpdate); }
-	else if (t == Problem_t::PB)  { return startPB(config, allowUpdate);  }
-	else if (t == Problem_t::ASP) { return startAsp(config, allowUpdate); }
-	else                          { throw std::domain_error("Unknown problem type!"); }
-}
-
-SatBuilder& ClaspFacade::startSat(ClaspConfig& config, bool allowUpdate) {
-	init(config, true);
-	initBuilder(new SatBuilder(config.enumerate.maxSat), allowUpdate);
-	return static_cast<SatBuilder&>(*builder_.get());
-}
-
-PBBuilder& ClaspFacade::startPB(ClaspConfig& config, bool allowUpdate) {
-	init(config, true);
-	initBuilder(new PBBuilder(), allowUpdate);
-	return static_cast<PBBuilder&>(*builder_.get());
-}
-
-Asp::LogicProgram& ClaspFacade::startAsp(ClaspConfig& config, bool allowUpdate) {
-	init(config, true);
-	Asp::LogicProgram* p = new Asp::LogicProgram();
-	lpStats_             = new Asp::LpStats;
-	p->accu              = lpStats_.get();
-	initBuilder(p, allowUpdate);
-	p->setOptions(config.asp);
-	p->setNonHcfConfiguration(config.testerConfig());
-	return *p;
-}
 ProgramBuilder& ClaspFacade::update(bool reloadConfig) {
 	CLASP_ASSERT_CONTRACT(builder_.get() && !solving());
 	CLASP_ASSERT_CONTRACT_MSG(step_.result.signal != SIGINT, "Interrupt not handled!");
@@ -372,13 +370,13 @@ bool ClaspFacade::prepare(EnumMode enumMode) {
 	CLASP_ASSERT_CONTRACT(config_ && !solving());
 	bool ok = this->ok();
 	SharedMinimizeData* m = 0;
-	EnumOptions& en       = config_->enumerate;
+	EnumOptions& en       = config_->solve;
 	if (builder_.get() && (ok = builder_->endProgram()) == true) {
 		builder_->getAssumptions(assume_);
-		if ((m  = en.opt != MinimizeMode_t::ignore ? builder_->getMinimizeConstraint(&en.bound) : 0) != 0) {
-			ok = m->setMode(en.opt, en.bound);
-			if (en.opt == MinimizeMode_t::enumerate && en.bound.empty()) {
-				ctx.report(warning(Event::subsystem_facade, "opt-mode=enum: no bound given, optimize statement ignored"));
+		if ((m  = en.optMode != MinimizeMode_t::ignore ? builder_->getMinimizeConstraint(&en.optBound) : 0) != 0) {
+			ok = m->setMode(en.optMode, en.optBound);
+			if (en.optMode == MinimizeMode_t::enumerate && en.optBound.empty()) {
+				ctx.report(warning(Event::subsystem_facade, "opt-mode=enum: no bound given, optimize statement ignored."));
 			}
 		}
 	}
@@ -387,7 +385,8 @@ bool ClaspFacade::prepare(EnumMode enumMode) {
 		ok = (enumMode == enum_volatile || ctx.addUnary(ctx.stepLiteral()));
 		if (ok) { solve_->prepare(ctx, m ? m->share() : 0, en.numModels); }
 	}
-	if (!accu_.get()) { builder_ = 0; }
+	if      (!accu_.get())  { builder_ = 0; }
+	else if (lpStats_.get()){ static_cast<Asp::LogicProgram*>(builder_.get())->dispose(false); }
 	return (ok && ctx.endInit()) || (stopStep(0, true), false);
 }
 void ClaspFacade::assume(Literal p) {
@@ -484,7 +483,7 @@ ExpectedQuantity::operator double() const { return valid() ? rep : std::numeric_
 ExpectedQuantity ClaspFacade::getStatImpl(const char* path, bool keys) const {
 #define GET_KEYS(o, path) ( ExpectedQuantity((o).keys(path)) )
 #define GET_OBJ(o, path)  ( ExpectedQuantity((o)[path]) )
-#define COMMON_KEYS "ctx.\0solvers.\0solver.\0costs.\0totalTime\0cpuTime\0solveTime\0unsatTime\0satTime\0numEnum\0optimal\0step\0result\0"
+#define COMMON_KEYS "ctx.\0solvers.\0solver.\0costs.\0time_total\0time_cpu\0time_solve\0time_unsat\0time_sat\0enumerated\0optimal\0step\0result\0"
 	static const char* _keys[] = {"accu.\0hccs.\0hcc.\0lp.\0" COMMON_KEYS, 0, "accu.\0lp.\0" COMMON_KEYS, "accu.\0" COMMON_KEYS };
 	enum ObjId { id_hccs, id_hcc, id_lp, id_ctx, id_solvers, id_solver, id_costs, id_total, id_cpu, id_solve, id_unsat, id_sat, id_num, id_opt, id_step, id_result };
 	if (!path)  path = "";
@@ -507,7 +506,7 @@ ExpectedQuantity ClaspFacade::getStatImpl(const char* path, bool keys) const {
 					if (oId == id_costs) {
 						char* x;
 						uint32 n = (uint32)std::strtoul(path, &x, 10);
-						uint32 N = stats->model() && stats->optimize() ? stats->costs()->numRules() : 0u;
+						uint32 N = stats->optimize() && stats->costs() ? (uint32)stats->costs()->size() : 0u;
 						if (x == path) {
 							if (!*path) { return keys ? ExpectedQuantity("__len\0") : ExpectedQuantity::error_ambiguous_quantity; }
 							if (!keys && std::strcmp(path, "__len") == 0) { return ExpectedQuantity(N); }
@@ -516,7 +515,7 @@ ExpectedQuantity ClaspFacade::getStatImpl(const char* path, bool keys) const {
 						if (*(path = x) == '.') { ++path; }
 						if (*path)  { return ExpectedQuantity::error_unknown_quantity; }
 						if (n >= N) { return ExpectedQuantity::error_not_available;    }
-						if (!keys)  { return ExpectedQuantity((double)stats->costs()->optimum(n)); }
+						if (!keys)  { return ExpectedQuantity((double)stats->costs()->at(n)); }
 					}
 					if (keys)            { return ExpectedQuantity(uint32(0)); }
 					if (oId <= id_sat)   { return *((&stats->totalTime) + (oId - id_total)); }
@@ -584,21 +583,25 @@ ExpectedQuantity ClaspFacade::getStat(const SharedContext& ctx, const char* key,
 	}
 	return res;	
 }
+Enumerator* ClaspFacade::enumerator() const { return solve_.get() ? solve_->enumerator() : 0; }
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspFacade::Summary
 /////////////////////////////////////////////////////////////////////////////////////////
 void ClaspFacade::Summary::init(ClaspFacade& f)  { std::memset(this, 0, sizeof(Summary)); facade = &f;}
 int ClaspFacade::Summary::stats()          const { return ctx().master()->stats.level(); }
 const Model* ClaspFacade::Summary::model() const { return facade->solve_.get() ? facade->solve_->lastModel() : 0; }
-const SharedMinimizeData* ClaspFacade::Summary::costs()const { return facade->solve_.get() ? facade->solve_->minimizer() : 0;  }
+bool ClaspFacade::Summary::optimize()      const {
+	if (const Enumerator* e = facade->enumerator()){
+		return e->optimize() || e->lastModel().opt;
+	}
+	return false;
+}
 const char* ClaspFacade::Summary::consequences() const {
 	int mt = facade->solve_.get() ? facade->solve_->modelType() : 0;
 	if ( (mt & CBConsequences::brave_consequences)    == CBConsequences::brave_consequences)   { return "Brave"; }
 	if ( (mt & CBConsequences::cautious_consequences) == CBConsequences::cautious_consequences){ return "Cautious"; }
 	return 0;
 }
-bool ClaspFacade::Summary::optimize()  const { return (facade->solve_.get() && facade->solve_->optimize()) || (model() && model()->opt); }
-bool ClaspFacade::Summary::optimum()   const { return (model() && model()->opt) || (costs() && complete()); }
 uint64 ClaspFacade::Summary::optimal() const { 
 	const Model* m = model();
 	if (m && m->opt) {
