@@ -17,11 +17,6 @@
 // along with Clasp; if not, write to the Free Software
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
-#if defined(_MSC_VER)
-#pragma warning (disable : 4146) // unary minus operator applied to unsigned type, result still unsigned
-#pragma warning (disable : 4996) // 'std::_Fill_n' was declared deprecated
-#endif
-
 #include <clasp/logic_program.h>
 #include <clasp/shared_context.h>
 #include <clasp/solver.h>
@@ -38,7 +33,7 @@ namespace Clasp { namespace Asp {
 /////////////////////////////////////////////////////////////////////////////////////////
 // class LpStats
 /////////////////////////////////////////////////////////////////////////////////////////
-#define LP_STATS( APPLY )                                      \
+#define LP_STATS( APPLY )                                        \
 	APPLY("bodies"              , bodies)                        \
 	APPLY("atoms"               , atoms)                         \
 	APPLY("atoms_aux"           , auxAtoms)                      \
@@ -185,9 +180,8 @@ void LogicProgram::setOptions(const AspOptions& opts) {
 	opts_ = opts;
 	if (opts.suppMod) {
 		if (!incData_ || incData_->startScc == 0) {
-			if (opts_.iters != 5 && ctx()) { ctx()->report(warning(Event::subsystem_prepare, "'supp-models' implies 'eq=0'.")); }
-			opts_.noEq();
-			opts_.noScc();
+			if (opts_.iters && ctx()) { ctx()->report(warning(Event::subsystem_prepare, "'supp-models' implies 'eq=0'.")); }
+			opts_.supportedModels();
 		}
 		else {
 			if (ctx()) { ctx()->report(warning(Event::subsystem_facade, "'supp-models' ignored for non-tight programs.")); }
@@ -440,6 +434,12 @@ LogicProgram& LogicProgram::unfreeze(Var atomId) {
 	return *this;
 }
 
+bool LogicProgram::isExternal(Var aId) const {
+	if (!aId || aId > numAtoms()) { return false; }
+	PrgAtom* a = getAtom(getEqAtom(aId));
+	return a->frozen() && (a->supports() == 0 || frozen());
+}
+
 LogicProgram& LogicProgram::addRule(const Rule& r) {
 	check_not_frozen();
 	// simplify rule
@@ -812,6 +812,12 @@ void LogicProgram::finalizeDisjunctions(Preprocessor& p, uint32 numSccs) {
 							*bIt           = WeightLiteral(negLit(h), 1);
 						}
 					}
+					else {
+						// only add support edge
+						for (PrgDisj::atom_iterator a = x->begin(), end = x->end(); a != end; ++a) {
+							B->addHead(getAtom(a->node()), PrgEdge::GAMMA_CHOICE_EDGE);
+						}
+					}
 				}
 			}
 		}
@@ -915,6 +921,16 @@ bool LogicProgram::addConstraints() {
 	if (!ctx()->ok() || !ctx()->addUnary(getFalseAtom()->trueLit())) {
 		return false;
 	}
+	if (options().noGamma && !disjunctions_.empty()) {
+		// add "rule" nogoods for disjunctions
+		for (DisjList::const_iterator it = disjunctions_.begin(); it != disjunctions_.end(); ++it) {
+			gc.start().add(~(*it)->literal());
+			for (PrgDisj::atom_iterator a = (*it)->begin(); a != (*it)->end(); ++a) {
+				gc.add(getAtom(a->node())->literal());
+			}
+			if (!gc.end()) { return false; }
+		}
+	}
 	// add bodies from this step
 	for (BodyList::const_iterator it = bodies_.begin(); it != bodies_.end(); ++it) {
 		if (!toConstraint((*it), *this, gc)) { return false; }
@@ -1034,7 +1050,7 @@ RuleType LogicProgram::simplifyBody(const Rule& r, BodyInfo& info) {
 			ruleState_.addToBody(lit);    // add to simplified body
 			sBody.push_back(WeightLiteral(lit, w));
 			pos += !lit.sign();
-			hash+= !lit.sign() ? hashId(lit.var()) : hashId(-lit.var());
+			hash+= hashLit(lit);
 		}
 		else if (!r.bodyIsSet()) {      // Merge duplicate lits
 			WeightLiteral& oldLit = info[info.findLit(lit)];
@@ -1047,6 +1063,9 @@ RuleType LogicProgram::simplifyBody(const Rule& r, BodyInfo& info) {
 				resType = WEIGHTRULE;
 			}
 		}
+		else {
+			bound -= 1;
+		}
 		dirty |= ruleState_.inBody(~lit);
 	}
 	weight_t minW  = 1;
