@@ -18,10 +18,8 @@
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
 #ifdef _MSC_VER
-#pragma warning (disable : 4146) // unary minus operator applied to unsigned type, result still unsigned
 #pragma warning (disable : 4996) // 'std::_Fill_n' was declared deprecated
 #endif
-
 #include <clasp/logic_program_types.h>
 #include <clasp/logic_program.h>
 #include <clasp/solver.h>
@@ -907,7 +905,7 @@ bool PrgBody::simplifyBody(LogicProgram& prg, bool strong, uint32* eqId) {
 	for (Literal* it = j, *end = j + size(); it != end; ++it) {
 		a       = it->var();
 		isEq    = a != prg.getEqAtom(a);
-		oldHash+= hashId(it->sign()?-a:a);
+		oldHash+= hashLit(*it);
 		if (isEq) {
 			prg.getAtom(a)->removeDep(id(), !it->sign()); // remove old edge
 			*it = prg.getAtom(a)->eqGoal(it->sign());     // replace with eq goal
@@ -971,14 +969,14 @@ bool PrgBody::simplifyBody(LogicProgram& prg, bool strong, uint32* eqId) {
 	uint32 newHash = 0;
 	weight_t sumW  = 0, reachW = 0;
 	for (uint32 p = 0, n = size_, i, h; p < n;) {
-		if      (!lits[p].sign())      { h = hashId( (a = lits[i = p++].var()));  }
-		else if (lits[n-1].sign())     { h = hashId(-(a = lits[i = --n].var()));  }
+		if      (!lits[p].sign())      { h = hashLit(lits[i = p++]);  }
+		else if (lits[n-1].sign())     { h = hashLit(lits[i = --n]);  }
 		else /* restore pos|neg order */ {
 			std::swap(lits[p], lits[n-1]);
 			if (jw) { std::swap(jw[p], jw[n-1]); }
 			continue;
 		}
-		assert(a == lits[i].var());
+		a = lits[i].var();
 		if (todos && todo.inBody(lits[i])) {
 			prg.getAtom(a)->addDep(id(), !lits[i].sign());
 			todo.clearBody(lits[i]);
