@@ -57,6 +57,7 @@ class DecisionHeuristicTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testDomInc);
 	CPPUNIT_TEST(testDomIncPrio);
 	CPPUNIT_TEST(testDomReinit);
+	CPPUNIT_TEST(testDomMinBug);
 	CPPUNIT_TEST_SUITE_END();
 public:
 	void testTrivial() {
@@ -635,6 +636,24 @@ public:
 		dom->select(s);
 		CPPUNIT_ASSERT(s.value(s.sharedContext()->symbolTable()[2].lit.var()) != value_free);
 	}
+
+	void testDomMinBug() {
+		SharedContext ctx;
+		ctx.master()->setHeuristic(new DomainHeuristic);
+		Solver& s = *ctx.master();
+		
+		LogicProgram api;
+		api.start(ctx).setAtomName(1, "a").setAtomName(2, "b")
+			.startRule().addHead(1).addToBody(2, false).endRule()
+			.startRule().addHead(2).addToBody(1, false).endRule()
+		;
+		addDomRule(api, "_heuristic(a,false,1)");
+		addDomRule(api, "_heuristic(b,false,1)");
+		
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+
+		CPPUNIT_ASSERT(ctx.symbolTable().domLits.size() == 2);
+	}
 	void addDomRule(LogicProgram& prg, const char* heu, Literal pre = posLit(0)) {
 		Var h = prg.newAtom();
 		prg.setAtomName(h, heu);
