@@ -23,6 +23,7 @@
 #include <clasp/clasp_facade.h>
 #include <clasp/minimize_constraint.h>
 #include <clasp/heuristics.h>
+#include <clasp/lookahead.h>
 
 namespace Clasp { namespace Test {
 using namespace Clasp::mt;
@@ -36,6 +37,12 @@ class FacadeTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testUpdateConfig);
 	CPPUNIT_TEST(testIncrementalProjectUpdate);
 	CPPUNIT_TEST(testIncrementalConfigUpdateBug);
+	CPPUNIT_TEST(testIncrementalLookaheadAddHeuristic);
+	CPPUNIT_TEST(testIncrementalDisableLookahead);
+	CPPUNIT_TEST(testIncrementalChangeLookahead);
+	CPPUNIT_TEST(testIncrementalExtendLookahead);
+	CPPUNIT_TEST(testIncrementalAddSolver);
+	CPPUNIT_TEST(testIncrementalRemoveSolver);
 	CPPUNIT_TEST(testStats);
 #if WITH_THREADS
 	CPPUNIT_TEST(testShareModeRegression);
@@ -102,8 +109,8 @@ public:
 	void testIncrementalEnum() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
-		config.enumerate.numModels = 0;
-		config.enumerate.type = EnumOptions::enum_record;
+		config.solve.numModels = 0;
+		config.solve.enumMode  = EnumOptions::enum_record;
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a");
 		asp.startRule(Clasp::Asp::CHOICERULE).addHead(1).endRule();
@@ -120,14 +127,14 @@ public:
 	void testIncrementalCons() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
-		config.enumerate.numModels = 0;
-		config.enumerate.type = EnumOptions::enum_cautious;
+		config.solve.numModels = 0;
+		config.solve.enumMode  = EnumOptions::enum_cautious;
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c");
 		asp.startRule(Clasp::Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).endRule();
 		libclasp.prepare();
 		CPPUNIT_ASSERT(libclasp.solve().sat());
-		config.enumerate.type = EnumOptions::enum_brave;
+		config.solve.enumMode = EnumOptions::enum_brave;
 		libclasp.update(true);
 		libclasp.prepare();
 		CPPUNIT_ASSERT(libclasp.solve().sat());
@@ -136,8 +143,8 @@ public:
 	void testIncrementalMin() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
-		config.enumerate.numModels = 0;
-		config.enumerate.type = EnumOptions::enum_auto;
+		config.solve.numModels = 0;
+		config.solve.enumMode  = EnumOptions::enum_auto;
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c");
 		asp.startRule(Clasp::Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).endRule();
@@ -153,8 +160,8 @@ public:
 	void testUpdateConfig() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
-		config.enumerate.numModels = 0;
-		config.enumerate.type = EnumOptions::enum_auto;
+		config.solve.numModels = 0;
+		config.solve.enumMode  = EnumOptions::enum_auto;
 		config.addSolver(0).heuId  = Heuristic_t::heu_berkmin;
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c");
@@ -165,14 +172,14 @@ public:
 		libclasp.update(true);
 		libclasp.prepare();
 		CPPUNIT_ASSERT(libclasp.solve().sat());
-		CPPUNIT_ASSERT(dynamic_cast<ClaspVsids*>(libclasp.ctx.master()->heuristic().get()));
+		CPPUNIT_ASSERT(dynamic_cast<ClaspVsids*>(libclasp.ctx.master()->heuristic()));
 	}
 	void testIncrementalProjectUpdate() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
-		config.enumerate.numModels = 0;
-		config.enumerate.type      = EnumOptions::enum_auto;
-		config.enumerate.project   = 1;
+		config.solve.numModels = 0;
+		config.solve.enumMode  = EnumOptions::enum_auto;
+		config.solve.project   = 1;
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "_a").setAtomName(2, "b");
 		asp.startRule(Clasp::Asp::CHOICERULE).addHead(1).addHead(2).endRule();
@@ -180,12 +187,12 @@ public:
 		CPPUNIT_ASSERT(libclasp.ctx.varInfo(libclasp.ctx.symbolTable()[2].lit.var()).project());
 		CPPUNIT_ASSERT(libclasp.solve().sat());
 		CPPUNIT_ASSERT(libclasp.summary().numEnum == 2);
-		config.enumerate.project = 0;
+		config.solve.project = 0;
 		libclasp.update(true);
 		libclasp.prepare();
 		CPPUNIT_ASSERT(libclasp.solve().sat());
 		CPPUNIT_ASSERT(libclasp.summary().numEnum == 4);
-		config.enumerate.project = 1;
+		config.solve.project = 1;
 		libclasp.update(true);
 		asp.setAtomName(3, "_x").setAtomName(4, "y");
 		asp.startRule(Clasp::Asp::CHOICERULE).addHead(3).addHead(4).endRule();
@@ -213,9 +220,105 @@ public:
 		libclasp.prepare();
 		CPPUNIT_ASSERT(asp.stats.auxAtoms == 0);
 	}
+	
+	void testIncrementalLookaheadAddHeuristic() {
+		Clasp::ClaspFacade libclasp;
+		Clasp::ClaspConfig config;
+		config.addSolver(0).lookType = Lookahead::atom_lookahead;
+		libclasp.startAsp(config, true).setAtomName(1, "a").setAtomName(2, "b").startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		libclasp.prepare();
+		PostPropagator* look = libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look);
+		CPPUNIT_ASSERT(look && look->next == 0);
+		config.addSolver(0).heuId = Heuristic_t::heu_unit;
+		libclasp.update(true);
+		libclasp.prepare();
+		look = libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look);
+		CPPUNIT_ASSERT(look && look->next == 0);
+	}
+	void testIncrementalDisableLookahead() {
+		Clasp::ClaspFacade libclasp;
+		Clasp::ClaspConfig config;
+		config.addSolver(0).lookType = Lookahead::atom_lookahead;
+		libclasp.startAsp(config, true).setAtomName(1, "a").setAtomName(2, "b").startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look));
+		config.addSolver(0).lookType = Lookahead::no_lookahead;
+		libclasp.update(true);
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look) == 0);
+	}
+	void testIncrementalChangeLookahead() {
+		Clasp::ClaspFacade libclasp;
+		Clasp::ClaspConfig config;
+		config.addSolver(0).lookType = Lookahead::atom_lookahead;
+		libclasp.startAsp(config, true).setAtomName(1, "a").setAtomName(2, "b").startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look));
+		config.addSolver(0).lookType = Lookahead::body_lookahead;
+		libclasp.update(true);
+		libclasp.prepare();
+		Lookahead* look = static_cast<Lookahead*>(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look));
+		CPPUNIT_ASSERT(look && look->score.types == Var_t::body_var );
+	}
+	void testIncrementalExtendLookahead() {
+		Clasp::ClaspFacade libclasp;
+		Clasp::ClaspConfig config;
+		config.addSolver(0).lookType = Lookahead::atom_lookahead;
+		config.addSolver(0).lookOps  = 3;
+		libclasp.startAsp(config, true).setAtomName(1, "a").setAtomName(2, "b").startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look));
+		config.addSolver(0).lookOps  = 0;
+		static_cast<Asp::LogicProgram&>(libclasp.update(true)).startRule(Asp::CHOICERULE).addHead(3).addHead(4).endRule();
+		libclasp.prepare();
+		Lookahead* look = static_cast<Lookahead*>(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look));
+		CPPUNIT_ASSERT(look && look->next == 0 );
+		while (libclasp.ctx.master()->numFreeVars() != 0) {
+			libclasp.ctx.master()->decideNextBranch();
+			libclasp.ctx.master()->propagate();
+			CPPUNIT_ASSERT(libclasp.ctx.master()->getPost(PostPropagator::priority_reserved_look) == look);
+		}
+	}
+	void testIncrementalAddSolver() {
+		Clasp::ClaspFacade libclasp;
+		Clasp::ClaspConfig config;
+		libclasp.startAsp(config, true).setAtomName(1, "a").setAtomName(2, "b").startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		libclasp.prepare();
+		CPPUNIT_ASSERT(!isSentinel(libclasp.ctx.stepLiteral()));
+		config.solve.setSolvers(2);
+		static_cast<Asp::LogicProgram&>(libclasp.update(true)).startRule(Asp::CHOICERULE).addHead(3).addHead(4).endRule();
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.ctx.concurrency() == 2 && libclasp.ctx.hasSolver(1));
+	}
+	
+	void testIncrementalRemoveSolver() {
+		Clasp::ClaspFacade libclasp;
+		Clasp::ClaspConfig config;
+		config.solve.numModels = 0;
+		config.solve.setSolvers(4);
+		libclasp.startAsp(config, true)
+			.setAtomName(1, "a").setAtomName(2, "b")
+			.startRule(Asp::CHOICERULE).addHead(1).addHead(2).addHead(4).addHead(3).endRule()
+			.startRule(Asp::CONSTRAINTRULE, 3).addHead(7).addToBody(1, true).addToBody(2, true).addToBody(3, true).addToBody(4, true).endRule()
+			.setCompute(7, false);
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.solve().sat());
+		CPPUNIT_ASSERT_EQUAL(uint64(11), libclasp.summary().numEnum);
+
+		config.solve.setSolvers(1);
+		static_cast<Asp::LogicProgram&>(libclasp.update(true)).startRule().addHead(7).addToBody(1, true).addToBody(2, true).endRule();
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.solve().sat() && libclasp.summary().numEnum == 10);
+
+		config.solve.setSolvers(2);
+		libclasp.update(true);
+		libclasp.prepare();
+		CPPUNIT_ASSERT(libclasp.solve().sat() && libclasp.summary().numEnum == 10);
+	}
 	void testStats() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
+		config.stats = 2;
 		Clasp::Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c");
 		asp.startRule(Clasp::Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).endRule();
@@ -229,9 +332,9 @@ public:
 			CPPUNIT_ASSERT(libclasp.getStat(it->c_str()).valid());
 		}
 		CPPUNIT_ASSERT(libclasp.getStat("lp.rules") == double(asp.stats.rules()));
-		CPPUNIT_ASSERT(libclasp.getStat("numEnum") == double(libclasp.summary().enumerated()));
+		CPPUNIT_ASSERT(libclasp.getStat("enumerated") == double(libclasp.summary().enumerated()));
 		CPPUNIT_ASSERT(libclasp.getStat("solvers.choices") == double(libclasp.ctx.master()->stats.choices));
-		CPPUNIT_ASSERT(libclasp.getStat("costs.0") == double(libclasp.summary().costs()->optimum(0)));
+		CPPUNIT_ASSERT(libclasp.getStat("costs.0") == double(libclasp.summary().costs()->at(0)));
 		CPPUNIT_ASSERT(libclasp.getStat("optimal") == double(libclasp.summary().optimal()));
 	}
 	void getKeys(const ClaspFacade& f, std::vector<std::string>& out, const std::string& p) {
@@ -249,7 +352,6 @@ public:
 		}
 	}
 #if WITH_THREADS
-	typedef ClaspFacade::AsyncResult AsyncResult;
 	void testShareModeRegression() {
 		Clasp::ClaspFacade libclasp;
 		Clasp::ClaspConfig config;
@@ -262,10 +364,11 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, libclasp.ctx.physicalShare(Constraint_t::static_constraint));
 		CPPUNIT_ASSERT_EQUAL(true, libclasp.ctx.physicalShare(Constraint_t::learnt_conflict));
 	}
+	typedef ClaspFacade::AsyncResult AsyncResult;
 	void testAsyncSolveTrivialUnsat() {
 		ClaspConfig config;
 		ClaspFacade libclasp;
-		config.enumerate.numModels = 0;
+		config.solve.numModels = 0;
 		Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a");
 		asp.startRule().addHead(1).addToBody(1, false).endRule();
@@ -277,7 +380,7 @@ public:
 	void testInterruptBeforeSolve() {
 		ClaspConfig config;
 		ClaspFacade libclasp;
-		config.enumerate.numModels = 0;
+		config.solve.numModels = 0;
 		Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a");
 		asp.startRule(Asp::CHOICERULE).addHead(1).endRule();
@@ -290,7 +393,7 @@ public:
 	void testCancelAsyncOperation() {
 		ClaspConfig config;
 		ClaspFacade libclasp;
-		config.enumerate.numModels = 0;
+		config.solve.numModels = 0;
 		Asp::LogicProgram& asp = libclasp.startAsp(config, true);
 		asp.setAtomName(1, "a");
 		asp.startRule(Asp::CHOICERULE).addHead(1).endRule();
